@@ -1,7 +1,7 @@
 import lightning as L
 from spotpython.data.lightdatamodule import LightDataModule, PadSequenceManyToMany
 from spotpython.utils.eda import generate_config_id
-from spotpython.utils.metrics import calculate_xai_consistency
+from spotpython.utils.metrics import calculate_xai_consistency_corr, calculate_xai_consistency_cosine, calculate_xai_consistency_euclidean, calculate_xai_consistency_spearman
 from pytorch_lightning.loggers import TensorBoardLogger
 from lightning.pytorch.callbacks.early_stopping import EarlyStopping
 from lightning.pytorch.callbacks import ModelCheckpoint
@@ -653,24 +653,31 @@ def train_model_xai(config: dict, fun_control: dict, timestamp: bool = True) ->
     # Perform feature attribution analysis
     model = trainer.model
     print("MODEL :", model)
-    model.eval()
 
-    # Get the validation dataloader from the LightningDataModule
+    # Get the validation and train dataloader from the LightningDataModule
+    train_dataloader: DataLoader = dm.train_dataloader()  # Fetch train data loader
     val_dataloader: DataLoader = dm.val_dataloader()  # Fetch validation data loader
 
     # Collect all validation data
     X_val_list = []
     y_val_list = []
 
+    # Collect all train data
+    X_train_list = []
+    y_train_list = []
+
+    # Iterate over the train dataloader to gather all data
+    for batch in train_dataloader:
+        X_batch, y_batch = batch  # Extract inputs and labels
+        X_train_list.append(X_batch)
+        y_train_list.append(y_batch)
+
     # Iterate over the validation dataloader to gather all data
     for batch in val_dataloader:
         X_batch, y_batch = batch  # Extract inputs and labels
         X_val_list.append(X_batch)
         y_val_list.append(y_batch)
 
-    # Concatenate all batches into single tensors
-    X_val_tensor = torch.cat(X_val_list, dim=0).to(model.device)
-
     # Perform feature attribution analysis
 
     # Check if at least 2 elements are in list fun_control["xai_methods"]
@@ -683,49 +690,76 @@ def train_model_xai(config: dict, fun_control: dict, timestamp: bool = True) ->
         if method not in valid_xai_methods:
             raise ValueError(f"Invalid XAI method: {method}. Valid methods are: {valid_xai_methods}")
 
+    X_train_tensor = torch.cat(X_train_list, dim=0).to(model.device)
+    X_train_tensor.requires_grad_()
+    X_val_tensor = torch.cat(X_val_list, dim=0).to(model.device)
+    X_val_tensor.requires_grad_()
+
     # Dictionary to store attributions
     attributions_dict = {}
 
     if fun_control["xai_baseline"] is None:
-        fun_control["xai_baseline"] = torch.zeros_like(X_val_tensor)
-        print("Baseline is None. Using zeros as baseline.")
+        X_train_mean = X_train_tensor.mean(dim=0)
+        fun_control["xai_baseline"] = X_train_mean.unsqueeze(0)
+        print("Baseline is None. Using training mean as baseline.")
     baseline = fun_control["xai_baseline"]
 
-    if "IntegratedGradients" in fun_control["xai_methods"]:
-        attr_ig = IntegratedGradients(model)
-        attribution_ig = attr_ig.attribute(X_val_tensor, baselines=baseline)
-        ig_attr_test_sum = attribution_ig.detach().numpy().sum(0)
-        row_sum_ig = np.sum(ig_attr_test_sum, axis=0)
-        if row_sum_ig == 0:
-            row_sum_ig += 1e-10
-        scaled_attribution_ig = ig_attr_test_sum / row_sum_ig
-        attributions_dict["IntegratedGradients"] = scaled_attribution_ig
+    # Use a subset of the validation data for attribution
+    if fun_control["xai_subset_size"] is not None:
+        N_total = X_val_tensor.size(0)
+        N_attr = min(fun_control["xai_subset_size"], N_total)
+        print(f"Using a subset of {N_attr} samples for attribution analysis out of {N_total} total samples.")
+        g = torch.Generator(device=X_val_tensor.device)
+        g.manual_seed(fun_control["seed"])
+        perm = torch.randperm(N_total, generator=g, device=X_val_tensor.device)[:N_attr]
+        X_val_tensor = X_val_tensor[perm]
+
+    # Ensure the model is in evaluation mode
+    model.eval()
+
+    target = fun_control.get("xai_target", None)
 
     if "KernelShap" in fun_control["xai_methods"]:
         attr_ks = KernelShap(model)
-        attribution_ks = attr_ks.attribute(X_val_tensor, baselines=baseline)
-        ks_attr_test_sum = attribution_ks.detach().numpy().sum(0)
-        row_sum_ks = np.sum(ks_attr_test_sum, axis=0)
-        if row_sum_ks == 0:
-            row_sum_ks += 1e-10
-        scaled_attribution_ks = ks_attr_test_sum / row_sum_ks
-        attributions_dict["KernelShap"] = scaled_attribution_ks
+        n_features = X_val_tensor.shape[1]
+        samples_ks = min(2000, 100 * n_features)  # Adjust number of samples based on features, maximum 2000
+        print("KernelShap: Using", samples_ks, "samples for attribution.")
+        with torch.no_grad():
+            attribution_ks = attr_ks.attribute(X_val_tensor, baselines=baseline, n_samples=samples_ks, perturbations_per_eval=64, target=target)
+        ks_attr_test_sum = attribution_ks.detach().numpy().sum(axis=0)
+        l2_norm = np.linalg.norm(ks_attr_test_sum)
+        l2_normalized_ks = ks_attr_test_sum / l2_norm if l2_norm != 0 else ks_attr_test_sum
+        attributions_dict["KernelShap"] = l2_normalized_ks
 
-    if "DeepLift" in fun_control["xai_methods"]:
-        attr_dl = DeepLift(model)
-        attribution_dl = attr_dl.attribute(X_val_tensor, baselines=baseline)
-        dl_attr_test_sum = attribution_dl.detach().numpy().sum(0)
-        row_sum_dl = np.sum(dl_attr_test_sum, axis=0)
-        if row_sum_dl == 0:
-            row_sum_dl += 1e-10
-        scaled_attribution_dl = dl_attr_test_sum / row_sum_dl
-        attributions_dict["DeepLift"] = scaled_attribution_dl
+    with torch.enable_grad():
+        if "IntegratedGradients" in fun_control["xai_methods"]:
+            attr_ig = IntegratedGradients(model)
+            attribution_ig = attr_ig.attribute(X_val_tensor, baselines=baseline, target=target)
+            vec = attribution_ig.detach().cpu().numpy().sum(axis=0)
+            l2 = np.linalg.norm(vec)
+            attributions_dict["IntegratedGradients"] = vec / l2 if l2 != 0 else vec
+
+        if "DeepLift" in fun_control["xai_methods"]:
+            attr_dl = DeepLift(model)
+            attribution_dl = attr_dl.attribute(X_val_tensor, baselines=baseline, target=target)
+            dl_attr_test_sum = attribution_dl.detach().numpy().sum(axis=0)
+            l2_norm = np.linalg.norm(dl_attr_test_sum)
+            l2_normalized_dl = dl_attr_test_sum / l2_norm if l2_norm != 0 else dl_attr_test_sum
+            attributions_dict["DeepLift"] = l2_normalized_dl
 
     attributions_list = [attributions_dict[method] for method in fun_control["xai_methods"]]
     attributions = np.stack(attributions_list, axis=0)
 
-    result_xai = calculate_xai_consistency(attributions)
-
-    # -------------------------------------------------------------------------------------------------------------------
+    # Calculate corr:
+    if fun_control["xai_metric"] not in ["corr", "cosine", "euclidean", "spearman"]:
+        raise ValueError(f"Invalid xai_metric: {fun_control['xai_metric']}. Valid metrics are: 'corr', 'cosine', 'euclidean', 'spearman'")
+    if fun_control["xai_metric"] == "corr":
+        result_xai = calculate_xai_consistency_corr(attributions)
+    elif fun_control["xai_metric"] == "cosine":
+        result_xai = calculate_xai_consistency_cosine(attributions)
+    elif fun_control["xai_metric"] == "euclidean":
+        result_xai = calculate_xai_consistency_euclidean(attributions)
+    elif fun_control["xai_metric"] == "spearman":
+        result_xai = calculate_xai_consistency_spearman(attributions)
 
     return result["val_loss"], result_xai
